/* custom keyboard functions

   Copyright (C) 2003 the Free Software Foundation

   Authors: 2003 Mateusz Golicz <mteg@jaszczur.org>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */



#include "config.h"
#include "edit.h"
#include "editcmddef.h"
#include <signal.h>
#include <ctype.h>

#define UNIKEY_SHIFT 0x00010000
#define UNIKEY_LALT   0x00020000
#define UNIKEY_RALT   0x00040000
#define UNIKEY_CONTROL 0x00080000

#define UNIKEY_METAKEYS 4
#define UNIKEY_META1 0x00100000
#define UNIKEY_META2 0x00200000
#define UNIKEY_META3 0x00400000
#define UNIKEY_META4 0x00800000

#define CK_Meta1 8000
#define CK_Meta2 8001
#define CK_Meta3 8002
#define CK_Meta4 8003

#define isblank(c) ((c) == ' ' || (c) == '\t')

struct uk_commands
{
	unsigned char * literal;
	unsigned char * descr;
	long cmd;
};

struct uk_modifiers
{
	unsigned char * literal;
	unsigned char * descr;
	u_int32_t modifier;
};

struct uk_keys
{
	unsigned char * literal;
	unsigned char * descr;
	int32_t key;
};


static struct uk_commands uk_commands[] =
	{{NULL,		"CURSOR MOVEMENT",	CK_No_Command},
	 {"up",		"Cursor up",		CK_Up},
	 {"down",	"Cursor down",		CK_Down},
	 {"left",	"Cursor left",		CK_Left},
	 {"right",	"Cursor right",		CK_Right},
	 {"word_left",	"Cursor word left",	CK_Word_Left},
	 {"word_right",	"Cursor word right",	CK_Word_Right},
	 {"enter",	"Newline",		CK_Enter},
	 {"return",	"Forced newline",	CK_Return},
	 {"scroll_up",	"Scroll up",		CK_Scroll_Up},
	 {"scroll_down","Scroll down",		CK_Scroll_Down},
	 {"page_up",	"Page Up",		CK_Page_Up},
	 {"page_down",	"Page Down",		CK_Page_Down},
	 {"paragraph_up",	"Paragraph up",		CK_Paragraph_Up},
	 {"paragraph_down",	"Paragraph down",	CK_Paragraph_Down},
	 {"home",	"Beginning of line",	CK_Home},
	 {"end",	"End of line",		CK_End},
	 {"beginning_of_text",	"Beginning of file",	CK_Beginning_Of_Text},
	 {"end_of_text",	"End of file",		CK_End_Of_Text},
	 {"begin_page",	"Beginning of page",	CK_Begin_Page},
	 {"end_page",	"End of page",		CK_End_Page},
	 {"backspace",	"Backspace",		CK_BackSpace},
	 {"delete",	"Delete",		CK_Delete},
	 {"delete_word_left","Delete word on left",	CK_Delete_Word_Left},
	 {"delete_word_right","Delete word on right",	CK_Delete_Word_Right},
	 {"tab",	"Insert tab",		CK_Tab},
	 {"undo",	"Undo",			CK_Undo},
	 
	 {NULL,		"MULTIFILE COMMANDS",	CK_No_Command},
  	 {"new",		"New window",		CK_NewWindow},
  	 {"split",		"Split window",		CK_NewWindow_Split},
	 {"next_window",	"Next window",	CK_Toggle},
	 {"prev_window",	"Previous window",	CK_ToggleBackwards},
	 {"size_plus",		"Enlarge window",	CK_SizePlus},
	 {"size_minus",		"Shrink window",	CK_SizeMinus},
	 {"fullscreen",		"Toggle fullscreen",	CK_Fullscreen},
	 {"next_window_fullscreen",	"Next window (fullscreen)",	CK_Toggle_Fullscreen},
	 {"prev_window_fullscreen",	"Previous window (fullscreen)",	CK_ToggleBackwards_Fullscreen},
	 {"quit",		"Quit",			CK_Shutdown},

	 {NULL,		"FILE COMMANDS",	CK_No_Command},
	 {"save",	"Save",			CK_Save},
	 {"load",	"Load",			CK_Load},
	 {"load_newwindow",	"Load in new window",			CK_Load_NewWindow},
	 {"clear",	"Clear",		CK_New},
	 {"save_as",	"Save as...",		CK_Save_As},
	 {"save_block",	"Save current block",	CK_Save_Block},
	 {"toggle_ro",	"Toggle read only",	CK_Toggle_ReadOnly},
	 
	 {NULL,		"BLOCK COMMANDS",	CK_No_Command},
	 {"mark",	"Toggle mark on/off",	CK_Mark},
	 {"unmark",	"Unmark block",		CK_Unmark},
	 {"copy",	"Copy block",		CK_Copy},
	 {"move",	"Move block",		CK_Move},
	 {"remove",	"Delete block",		CK_Remove},
	 {"find",	"Find",			CK_Find},
	 {"find_again",	"Find again",		CK_Find_Again},
	 {"replace",	"Replace",		CK_Replace},
	 {"replace_again","Replace again",	CK_Replace_Again},
	 
	 {NULL,		"MISCELLANEOUS COMMANDS",	CK_No_Command},
	 {"menu",	"Drop menu",			CK_Menu},
	 {"about",	"About",		CK_About},
	 {"insertliteral",	"Insert literal",		CK_InsertLiteral},
	 {"options",	"Options dialog",	CK_Options},
	 {"options_save",	"Save options dialog",	CK_Options_Save},
	 {"wrap",	"Word wrap",		CK_Wrap},
	 {"insert_file", "Insert file",		CK_Insert_File},
	 {"close",	"Close",		CK_Exit},
	 {"toggle_insert",	"Toggle ins/del",	CK_Toggle_Insert},
	 {"help",	"Show keys",		CK_Help},
	 {"date",	"Insert date",		CK_Date},
	 {"refresh",	"Refresh",		CK_Refresh},
	 {"goto",	"Goto line",		CK_Goto},
	 {"delete_line",	"Delete line",		CK_Delete_Line},
	 {"delete_to_line_end","Delete after cursor", CK_Delete_To_Line_End},
	 {"delete_to_line_begin","Delete before cursor", CK_Delete_To_Line_Begin},
	 {"sort",	"Sort block",		CK_Sort},
	 {"mail",	"Mail",			CK_Mail},
	 {"paragraph_format",	"Paragraph format",	CK_Paragraph_Format},
	 {"toggle_syntax",	"Toggle syntax checking", CK_Toggle_Syntax},
	 
	 {NULL,		"HIGHLIGHT COMMANDS",	CK_No_Command},
	 {"up_highlight",		"Cursor up w/highlight",		CK_Up_Highlight},
	 {"down_highlight",	"Cursor down w/highlight",		CK_Down_Highlight},
	 {"left_highlight",	"Cursor left w/highlight",		CK_Left_Highlight},
	 {"right_highlight",	"Cursor right w/highlight",		CK_Right_Highlight},
	 {"word_left_highlight",	"Cursor word left w/h",	CK_Word_Left_Highlight},
	 {"word_right_highlight",	"Cursor word right w/h",	CK_Word_Right_Highlight},
	 {"scroll_up_highlight",	"Scroll up w/h",		CK_Scroll_Up_Highlight},
	 {"scroll_down_highlight","Scroll down w/h",		CK_Scroll_Down_Highlight},
	 {"page_up_highlight",	"Page Up w/highlight",		CK_Page_Up_Highlight},
	 {"page_down_highlight",	"Page Down w/highlight",		CK_Page_Down_Highlight},
	 {"paragraph_up_highlight",	"Paragraph up w/h",		CK_Paragraph_Up_Highlight},
	 {"paragraph_down_highlight",	"Paragraph down w/h",	CK_Paragraph_Down_Highlight},
	 {"home_highlight",	"Beginning of line w/h",	CK_Home_Highlight},
	 {"end_highlight",	"End of line w/h",		CK_End_Highlight},
	 {"beginning_of_text_highlight",	"Beginning of file w/h",	CK_Beginning_Of_Text_Highlight},
	 {"end_of_text_highlight",	"End of file w/h",		CK_End_Of_Text_Highlight},
	 {"begin_page_highlight",	"Beginning of page w/h",	CK_Begin_Page_Highlight},
	 {"end_page_highlight",	"End of page w/h",		CK_End_Page},

	 {NULL,		"MACROS",	CK_No_Command},
	 {"begin_record_macro",	"Begin recording macro",	CK_Begin_Record_Macro},
	 {"end_record_macro",	"End recording macro",		CK_End_Record_Macro},
	 {"delete_macro",	"Delete macro",			CK_Delete_Macro},
	 {"execute_macro",	"Execute macro",		CK_Execute_Macro},
	 
	 {NULL,		"META KEYS",	CK_No_Command},
	 {"meta1",	"Meta key 1",		CK_Meta1},
	 {"meta2",	"Meta key 2",		CK_Meta2},
	 {"meta3",	"Meta key 3",		CK_Meta3},
	 {"meta4",	"Meta key 4",		CK_Meta4},
	 {NULL,	NULL,		0}};

static struct uk_modifiers	uk_modifiers[] = 
	{{"lalt",		"LALT", UNIKEY_LALT},
	 {"ralt",		"RALT", UNIKEY_RALT},
	 {"shift",		"SHFT", UNIKEY_SHIFT},
	 {"control",		"CTRL", UNIKEY_CONTROL},
	 {"meta1",		"META1", UNIKEY_META1},
	 {"meta2",		"META2", UNIKEY_META2},
	 {"meta3",		"META3", UNIKEY_META3},
	 {"meta4",		"META4", UNIKEY_META4},
	 {NULL,			NULL,	0}};

static struct uk_keys		uk_keys[] =
	{{"down",		"DOWN",		KEY_DOWN},
	 {"up",			"UP",		KEY_UP},
	 {"left",		"LEFT",		KEY_LEFT},
	 {"right",		"RIGHT",	KEY_RIGHT},
	 {"backspace",		"BS",		KEY_BACKSPACE},
	 {"f1",			"F1",		KEY_F (1)},
	 {"f2",			"F2",		KEY_F (2)},
	 {"f3",			"F3",		KEY_F (3)},
	 {"f4",			"F4",		KEY_F (4)},
	 {"f5",			"F5",		KEY_F (5)},
	 {"f6",			"F6",		KEY_F (6)},
	 {"f7",			"F7",		KEY_F (7)},
	 {"f8",			"F8",		KEY_F (8)},
	 {"f9",			"F9",		KEY_F (9)},
	 {"f10",		"F10",		KEY_F (10)},
	 {"f11",		"F11",		KEY_F (11)},
	 {"f12",		"F12",		KEY_F (12)},
	 {"f13",		"F13",		KEY_F (13)},
	 {"f14",		"F14",		KEY_F (14)},
	 {"f15",		"F15",		KEY_F (15)},
	 {"f16",		"F16",		KEY_F (16)},
	 {"f17",		"F17",		KEY_F (17)},
	 {"f18",		"F18",		KEY_F (18)},
	 {"f19",		"F19",		KEY_F (19)},
	 {"f20",		"F20",		KEY_F (20)},
	 {"f21",		"F21",		KEY_F (21)},
	 {"f22",		"F22",		KEY_F (22)},
	 {"f23",		"F23",		KEY_F (23)},
	 {"f24",		"F24",		KEY_F (24)},
	 {"del",		"DEL",		KEY_DC},
	 {"ins",		"INS",		KEY_IC},
	 {"pgup",		"PGUP",		KEY_PPAGE},
	 {"pgdn",		"PGDN",		KEY_NPAGE},
	 {"a1",			"KEYPAD7",	KEY_A1},
	 {"a3",			"KEYPAD9",	KEY_A3},
	 {"b2",			"KEYPAD5",	KEY_B2},
	 {"c1",			"KEYPAD1",	KEY_C1},
	 {"c3",			"KEYPAD3",	KEY_C3},
	 {"home",		"HOME",		KEY_HOME},
	 {"end",		"END",		KEY_END},
	
	 {NULL,		NULL,	0}};

u_int32_t unikey_translatecommand(unsigned char * cmd)
{
	int i;
	
	if(strncmp(cmd, "switch:", 7) == 0)
		return(CK_SwitchTo(atoi(cmd + 7)));

	if(strncmp(cmd, "fsswitch:", 9) == 0)
		return(CK_SwitchTo_Fullscreen(atoi(cmd + 9)));
	
	if(strncmp(cmd, "generator:", 10) == 0)
		return(register_pipeoper(cmd + 10, 0));

	if(strncmp(cmd, "passthru:", 9) == 0)
		return(register_pipeoper(cmd + 9, 1));
	
	for(i = 0; uk_commands[i].cmd; i++)
		if(uk_commands[i].literal)
			if(!strcmp(uk_commands[i].literal, cmd))
				return(uk_commands[i].cmd);
	return(0);
}

static int translate_line(unsigned char * line, int32_t * key, u_int32_t * action)
{
	unsigned char * str = strdup(line);
	unsigned char * s;
	unsigned char * p;
	int i, ready = 0;
	u_int32_t cmd = 0;
	int32_t decoded = 0;
	
	if((s = whitespace_lastindex(p = str)))
	{
		whitespace_trim(p);
		
		cmd = unikey_translatecommand(p);
	
		if(cmd)
		{
			while(s)
			{
				unsigned char * n = whitespace_lastindex(s);
				
				whitespace_trim(s);
//				fprintf(stderr, "parsing: '%s'\n", s);
				
				for(i = 0; uk_modifiers[i].literal; i++)
					if(!strcmp(uk_modifiers[i].literal, s))
					{
						decoded |= uk_modifiers[i].modifier;
						break;
					}
				
				if(ready)
				{
					fprintf(stderr, "Syntax error (bogus characters)\n");
					free(str);
					return(0);
				}
				if(! uk_modifiers[i].literal)
				{
					for(i = 0; uk_keys[i].literal; i++)
						if(!strcmp(uk_keys[i].literal, s))
						{
							decoded |= uk_keys[i].key;
							ready = 1;
							break;
						}

					if(! uk_keys[i].literal)
					{
						unsigned char c;
						
						c = *s;
						if(c == '#' || !s[1])
						{
							if(c == '#')
								c = atoi(s + 1);
						
							decoded |= c;
							ready = 1;
							if(decoded & (UNIKEY_LALT | UNIKEY_RALT))
								decoded |= 0x200;
						}
					}
				}
				
				s = n;
			}
			if(ready)
			{
				int32_t cdecoded = decoded & 0xff;
				if((decoded & UNIKEY_CONTROL) && ((cdecoded >= 'a' && cdecoded <= ('a' + 32)) || (cdecoded >= 'A' && cdecoded <= ('A' + 32)) || (cdecoded < ' ')))
					decoded &= 0xffffff1f & (~(UNIKEY_CONTROL));

				*action = cmd;
				*key = decoded;
				return(1);
			}
		}
		else
			fprintf(stderr, "No such command: '%s'\n", p);
	}
	else
		fprintf(stderr, "Syntax error.\n");
	
	free(str);
	return(0);
}

static int32_t * unikey_mapping = NULL;
const static int32_t unikey_permanent_mapping[] = {KEY_F(1), CK_NewWindow,
					     KEY_F(2), CK_Save,
					     KEY_F(12), CK_Save_As,
					     KEY_F(3), CK_Mark,
					     KEY_F(13), CK_Unmark,
					     KEY_F(4), CK_Replace,
					     KEY_F(14), CK_Replace_Again,
					     KEY_F(5), CK_Copy,
					     KEY_F(6), CK_Move,
					     KEY_F(7), CK_Find,
					     KEY_F(17), CK_Find_Again,
					     KEY_F(8), CK_Remove,
					     KEY_F(18), CK_New,
					     KEY_F(10), CK_Exit,
					     KEY_F(20), CK_Shutdown};

#define MCTRL(x) (((x) & 0x1f))

const static int32_t unikey_default_mapping[] = {
	KEY_BACKSPACE, CK_BackSpace,	KEY_DC,	CK_Delete,	0x7f, CK_BackSpace,
	MCTRL('q'), CK_InsertLiteral,	MCTRL('d'), CK_Delete,
	'\n',	CK_Enter, '\r', CK_Return,
	KEY_PPAGE, CK_Page_Up,		KEY_NPAGE, CK_Page_Down,
	KEY_LEFT, CK_Left,		KEY_RIGHT, CK_Right,
	KEY_UP,	CK_Up,			KEY_DOWN, CK_Down,
	UNIKEY_LALT | '\t', CK_Toggle,	
	UNIKEY_LALT | '\n', CK_Fullscreen,
	UNIKEY_SHIFT | '\n', CK_Return,
	KEY_HOME, CK_Home,		KEY_END, CK_End,
	'\t',	CK_Tab,			MCTRL('u'), CK_Undo,
	KEY_IC,	CK_Toggle_Insert,	MCTRL('s'), CK_Toggle_Syntax,
	MCTRL('o'), CK_Load,		MCTRL('n'), CK_NewWindow,
	MCTRL('l'), CK_Refresh,		ESC_CHAR, CK_Exit,
	KEY_F(15), CK_Insert_File,
	UNIKEY_LALT | 't', CK_Sort,	UNIKEY_LALT | 'm', CK_Mail,
	MCTRL('z'), CK_Word_Left,	MCTRL('x'), CK_Word_Right,
	MCTRL('h'), CK_BackSpace,	UNIKEY_LALT | 'l', CK_Goto,
	MCTRL('y'), CK_Delete_Line,	UNIKEY_LALT | 'p', CK_Paragraph_Format,
	UNIKEY_LALT | UNIKEY_SHIFT | KEY_PPAGE, CK_Beginning_Of_Text_Highlight,
	UNIKEY_LALT | UNIKEY_SHIFT | KEY_NPAGE, CK_End_Of_Text_Highlight,
	UNIKEY_LALT | UNIKEY_SHIFT | KEY_LEFT, CK_Word_Left_Highlight,
	UNIKEY_LALT | UNIKEY_SHIFT | KEY_RIGHT, CK_Word_Right_Highlight,
	UNIKEY_SHIFT | KEY_PPAGE, CK_Page_Up_Highlight,
	UNIKEY_SHIFT | KEY_NPAGE, CK_Page_Down_Highlight,
	UNIKEY_SHIFT | KEY_LEFT, CK_Left_Highlight,
	UNIKEY_SHIFT | KEY_RIGHT, CK_Right_Highlight,
	UNIKEY_SHIFT | KEY_UP, CK_Up_Highlight,
	UNIKEY_SHIFT | KEY_DOWN, CK_Down_Highlight,
	UNIKEY_SHIFT | KEY_HOME, CK_Home_Highlight,
	UNIKEY_SHIFT | KEY_END, CK_End_Highlight,
	UNIKEY_SHIFT | KEY_F(2), CK_Save_As,
	UNIKEY_SHIFT | KEY_F(3), CK_Unmark,
	UNIKEY_SHIFT | KEY_F(4), CK_Replace_Again,
	UNIKEY_SHIFT | KEY_F(7), CK_Find_Again,
	UNIKEY_LALT | KEY_BACKSPACE, CK_Undo,
	UNIKEY_LALT | KEY_PPAGE, CK_Beginning_Of_Text,
	UNIKEY_LALT | KEY_NPAGE, CK_End_Of_Text,
	UNIKEY_LALT | KEY_UP, CK_Scroll_Up,
	UNIKEY_LALT | KEY_DOWN, CK_Scroll_Down,
	UNIKEY_LALT | KEY_LEFT, CK_Word_Left,
	UNIKEY_LALT | KEY_RIGHT, CK_Word_Right,
	UNIKEY_LALT | KEY_DC, CK_Remove,
	KEY_F(1), CK_NewWindow,		KEY_F(2), CK_Save,
	KEY_F(12), CK_Save_As,
	UNIKEY_SHIFT | KEY_F(12), CK_Save_As,		KEY_F(3), CK_Mark,
	UNIKEY_SHIFT | KEY_F(13), CK_Unmark,		KEY_F(4), CK_Replace,
	UNIKEY_SHIFT | KEY_F(14), CK_Replace_Again,	KEY_F(5), CK_Copy,
	KEY_F(6), CK_Move,		KEY_F(7), CK_Find,
	UNIKEY_SHIFT | KEY_F(17), CK_Find_Again,	KEY_F(8), CK_Remove,
	UNIKEY_SHIFT | KEY_F(18), CK_New,		KEY_F(10), CK_Exit,
	UNIKEY_SHIFT | KEY_F(20), CK_Shutdown,
	UNIKEY_LALT | 'a', CK_SizeMinus, UNIKEY_LALT | 'z', CK_SizePlus, 
	MCTRL('a'), CK_Execute_Macro, 
	KEY_F(11), CK_Load_NewWindow, UNIKEY_SHIFT | KEY_F(11), CK_Load_NewWindow,
	MCTRL('c'), CK_Shutdown,
	UNIKEY_LALT | '1', CK_SwitchTo_Fullscreen(0),
	UNIKEY_LALT | '2', CK_SwitchTo_Fullscreen(1),
	UNIKEY_LALT | '3', CK_SwitchTo_Fullscreen(2),
	UNIKEY_LALT | '4', CK_SwitchTo_Fullscreen(3),
	UNIKEY_LALT | '5', CK_SwitchTo_Fullscreen(4),
	UNIKEY_LALT | '6', CK_SwitchTo_Fullscreen(5),
	UNIKEY_LALT | '7', CK_SwitchTo_Fullscreen(6),
	UNIKEY_LALT | '8', CK_SwitchTo_Fullscreen(7),
	UNIKEY_LALT | '9', CK_SwitchTo_Fullscreen(8),
	UNIKEY_LALT | '0', CK_SwitchTo_Fullscreen(9),
	MCTRL('r'), CK_Toggle_ReadOnly,
	MCTRL('f'), CK_Meta1,
	UNIKEY_META1 | '1', CK_NewWindow,
	UNIKEY_META1 | '2', CK_Save,
	UNIKEY_META1 | '3', CK_Mark,
	UNIKEY_META1 | '4', CK_Replace,
	UNIKEY_META1 | '5', CK_Copy,
	UNIKEY_META1 | '6', CK_Move,
	UNIKEY_META1 | '7', CK_Find,
	UNIKEY_META1 | '8', CK_Remove,
	UNIKEY_META1 | '9', CK_Menu,
	UNIKEY_META1 | '!', CK_Load_NewWindow,
	UNIKEY_META1 | '@', CK_Save_As,
	UNIKEY_META1 | '#', CK_Unmark,
	UNIKEY_META1 | '$', CK_Replace_Again,
	UNIKEY_META1 | '&', CK_Find_Again,
	UNIKEY_META1 | '*', CK_Remove,
	UNIKEY_META1 | ')', CK_Shutdown,

	0, 0};
					     
static unsigned char unikey_metakey_states[UNIKEY_METAKEYS];

int unikey_load_mapping(FILE *fh)
{
	unsigned char buffer[256];
	unsigned char * line;
	int lineno = 0;
	int transno = 0;

	free(unikey_mapping);
	if(!fh)
	{
		unikey_mapping = (int32_t*) malloc(sizeof(unikey_default_mapping));
		memcpy(unikey_mapping, unikey_default_mapping, sizeof(unikey_default_mapping));
		
		return(1);
	}

	unikey_mapping = (int32_t*) malloc(sizeof(unikey_permanent_mapping) + 2 * sizeof(int32_t));
	memcpy(unikey_mapping, unikey_permanent_mapping, sizeof(unikey_permanent_mapping));
	transno = sizeof(unikey_permanent_mapping) / sizeof(int32_t) / 2;
//	fprintf(stderr, "%d\n", transno);
	unikey_mapping[transno * 2] = 0;
	unikey_mapping[transno * 2 + 1] = 0;
	
	while((line = fgets(buffer, 255, fh)))
	{
		unsigned char *r;
		int32_t key;
		u_int32_t action;
		
		lineno ++;

		while(isblank(*line)) line++;
		

		if((r = index(line, '\n'))) *r = 0;
		if((r = index(line, '\r'))) *r = 0;
		if(*line == '#') continue;
		if(!whitespace_lastindex(line)) continue;
/*		if(strncmp(line, "disable signal ", 15) == 0)
		{
			int n = atoi(line + 15);
			signal(n, SIG_IGN);
			continue;
		}
*/	
		
		if(!translate_line(line, &key, &action))
		{
			fprintf(stderr, "Problems loading keymap file at line %d\n", lineno);
			fclose(fh);
			free(unikey_mapping);
			unikey_mapping = NULL;
			return(0);
		}
		
		unikey_mapping = (int32_t*) realloc(unikey_mapping, sizeof(int32_t) * 2 * (transno + 2));
		unikey_mapping[transno * 2] = key;
		unikey_mapping[transno * 2 + 1] = action;
		transno++;
		unikey_mapping[transno * 2] = 0;
		unikey_mapping[transno * 2 + 1] = 0;
	}

	memset(unikey_metakey_states, 0, sizeof(unsigned char) * UNIKEY_METAKEYS);
	
	fclose(fh);
	return(1);
}


void unikey_decode(u_int32_t x_state, int32_t x_key, int * command, int * charforins)
{
	int32_t current;
	u_int32_t rescmd = 0;
	int i = 0, cm = 0;
	
	current = x_key;
	if(x_state & SHIFT_PRESSED) current |= UNIKEY_SHIFT;

	switch(current)
	{
		case KEY_SLEFT: current = KEY_LEFT | UNIKEY_SHIFT; break;
		case KEY_SRIGHT: current = KEY_RIGHT | UNIKEY_SHIFT; break;
		case KEY_SHOME: current = KEY_HOME | UNIKEY_SHIFT; break;
		case KEY_SEND: current = KEY_END | UNIKEY_SHIFT; break;
	}

	if(x_state & ALTL_PRESSED) current = ALT(current);
	if(x_state & ALTR_PRESSED) current |= UNIKEY_RALT;

//	if(x_state & CONTROL_PRESSED) current |= UNIKEY_CONTROL;

	if((current & 0xff) < ' ')
		current &= ~UNIKEY_CONTROL;

	if(unikey_metakey_states[0]) current |= UNIKEY_META1;
	if(unikey_metakey_states[1]) current |= UNIKEY_META2;
	if(unikey_metakey_states[2]) current |= UNIKEY_META3;	
	if(unikey_metakey_states[3]) current |= UNIKEY_META4;

	if(unikey_mapping)
	{
		for(i = 0; unikey_mapping[i] || unikey_mapping[i + 1]; i += 2)
		{
			if(unikey_mapping[i] == current)
			{
				rescmd = unikey_mapping[i + 1];
				break;
			}
		}
	}

//	fprintf(stderr, "CMD: %x RES %x\n",current, rescmd);

	if(rescmd)
	{
		switch(rescmd)
		{
			case CK_Meta1: unikey_metakey_states[0] = 1; cm = 1; break;
			case CK_Meta2: unikey_metakey_states[1] = 1; cm = 1; break;
			case CK_Meta3: unikey_metakey_states[2] = 1; cm = 1; break;
			case CK_Meta4: unikey_metakey_states[3] = 1; cm = 1; break;
			case CK_InsertLiteral:
				*charforins = edit_raw_key_query (_(" Insert Literal "), _(" Press any key: "), 0);
				return;
			default:
				memset(unikey_metakey_states, 0, sizeof(unsigned char) * UNIKEY_METAKEYS);
				*command = rescmd;
				return;
		}
		return;
	}
	
//	fprintf(stderr, "CMD: %x (%d) (surrendered after %d tries, size %d %x)\n", current, KEY_F(12),i ,sizeof(unikey_default_mapping) /2/ sizeof(u_int32_t), ALTL_PRESSED);

	if (x_key < 256 && is_printable (x_key)) {

		*charforins = x_key;
	}
	if(!cm) memset(unikey_metakey_states, 0, sizeof(unsigned char) * UNIKEY_METAKEYS);
}

static unsigned char ** listaddstr(unsigned char ** list, int ni, unsigned char * string)
{
	list = (unsigned char**) realloc(list, (ni + 2) * sizeof(unsigned char*));
	list[ni] = string;
	list[ni + 1] = NULL;
	return(list);
}

static unsigned char * decodekey(int32_t key)
{
	unsigned char buffer[200];
	int i;
	buffer[0] = 0;
	for(i = 0; uk_modifiers[i].literal; i++)
	{
		if(key & uk_modifiers[i].modifier)
		{
			key &= ~uk_modifiers[i].modifier;
			sprintf(buffer, "%s %s", buffer, uk_modifiers[i].descr); 
		}
	}
	
	for(i = 0; uk_keys[i].literal; i++)
	{
		if(key == uk_keys[i].key)
		{
			sprintf(buffer, "%s %s", buffer, uk_keys[i].descr);
			return(strdup(buffer));
		}
	}
	
//	if(key < ' ') { key += 'A' - 1; }
	if(key < ' ')
	{
		key += 'A' - 1;
		sprintf(buffer, "%s ^%c", buffer, key);
	}
	else if(is_printable(key))
		sprintf(buffer, "%s %c", buffer, key);
	else
		sprintf(buffer, "%s #0x%x", buffer, key);
	
	return(strdup(buffer));
}

unsigned char ** unikey_create_help(void)
{
	unsigned char ** ret = NULL;
	unsigned char txbuffer[200];
	int i, ni = 0;
	
	for(i = 0; uk_commands[i].cmd; i++)
	{
		if(uk_commands[i].literal)
		{
			int j, first = 1;
			for(j = 0; unikey_mapping[j] || unikey_mapping[j + 1]; j += 2)
			{
				if(unikey_mapping[j + 1] == uk_commands[i].cmd)
				{
					unsigned char *k;
					int32_t key;
					
					key = unikey_mapping[j];
					k = decodekey(key);
					
					if((first--) > 0)
						snprintf(txbuffer, 200, "%-35s %-25s", uk_commands[i].descr, k);
					else
						snprintf(txbuffer, 200, "%-35s %-25s", "", k); 
					
					free(k);
					ret = listaddstr(ret, ni++, strdup(txbuffer));
				}
			}
		}
		else
		{
			ret = listaddstr(ret, ni++, strdup(""));
			ret = listaddstr(ret, ni++, strdup(uk_commands[i].descr));
		}
	}
	return(ret);
}

void unikey_dispose_help(unsigned char ** list)
{
        int i;
        if(!list) return;
        for(i = 0; list[i]; i++) free(list[i]);
        free(list);
        return;
}

